// audio library identification routines
#include <windows.h>
#include <stdio.h>
#include <math.h>
#include <float.h>
#include "main.h"
#include "memory.h"
#include "cpu.h"
#include "al.h"

typedef struct {
	char * name;    // name of the function
	char * libname; // library to which the function belongs
	char * sig;     // the function's signature
	DWORD address;
} alibfcn;

#define MAX_NUM_AL_FCNS 256
#define NUM_SIG_FILES 1

const char * libsigs[NUM_SIG_FILES] = {
	"c:\\bnkf.txt"
};


alibfcn AL_ARRAY[MAX_NUM_AL_FCNS];
int al_fcn_cnt = 0;

char * StrAllocCopy(const char * in) {
	char * temp = malloc(strlen(in)+1);
	strcpy(temp,in);
	return temp;
}

void LoadLibrarySignatures(void) {
	int c;

	LoadFcnStr("", "alBnkfNew","27bdffd0afbf002cafb40028afb30024afb20020afb1001cafb00018848e0000");
	//LoadFcnStr("", "alBnkfNew", "27bdffd0afbf002cafb40028afb30024afb20020afb1001cafb00018848e0000240142310080802515c1002e00a08825848f0002000098250080902519e00029");

#if 0
    for (c=0; c < NUM_SIG_FILES; c++) {
		FILE * libtab;
		libtab = fopen(libsigs[c],"r");
		if (!libtab) {
			DisplayError("failed to open %s", libsigs[c]);
			continue;
		}

		//LoadFcn(libtab,"osAiSetFrequency",libsigs[c]);
		//LoadFcn(libtab,"alCSeqNew",libsigs[c]);
		//LoadFcn(libtab,"alSeqFileNew",libsigs[c]);
		LoadFcn(libtab,"alBnkfNew",libsigs[c]);
		//LoadFcn(libtab,"alInit",libsigs[c]);
		//LoadFcn(libtab,"alCSPNew",libsigs[c]);
		//LoadFcn(libtab,"alCSPSendMidi",libsigs[c]);

		fclose(libtab);
	}
#endif
}

void LoadFcn(FILE * libtab, const char * fcnname, const char * libname) {
	char buffer1[8192*2+1],buffer2[8192*2+1];
	rewind(libtab);
	do {
		fscanf(libtab,":%s\n%s\n",buffer1,buffer2);
		if (feof(libtab)) break;
		if (!strcmp(buffer1,fcnname)) {
			AL_ARRAY[al_fcn_cnt].name=StrAllocCopy(buffer1);
			AL_ARRAY[al_fcn_cnt].libname=StrAllocCopy(libname);
			AL_ARRAY[al_fcn_cnt].sig=StrAllocCopy(buffer2);
			AL_ARRAY[al_fcn_cnt].address=0;
			al_fcn_cnt++;
			//DisplayError("loaded %s (%s)",buffer1,libname);
		}
			
	} while (!feof(libtab));
}

void LoadFcnStr(const char * libname, const char * fcnname, const char * fcnsig)
{
	AL_ARRAY[al_fcn_cnt].name=StrAllocCopy(fcnname);
	AL_ARRAY[al_fcn_cnt].libname=StrAllocCopy(libname);
	AL_ARRAY[al_fcn_cnt].sig=StrAllocCopy(fcnsig);
	AL_ARRAY[al_fcn_cnt].address=0;
	al_fcn_cnt++;
}


void alIdentify(DWORD PC) {
	int c;
	DWORD i;
	BYTE v;
	unsigned int lv;
	
	for (c=0; c < al_fcn_cnt; c++) {
		unsigned int len = strlen(AL_ARRAY[c].sig);
		for (i=PC; (i-PC)*2 < len; i++) {
			if (AL_ARRAY[c].sig[(i-PC)*2]=='.') continue;
			sscanf(AL_ARRAY[c].sig+((i-PC)*2),"%02x",&lv);
			r4300i_LB_VAddr(i,&v);
			if (v != lv) break;
		}
		if ((i-PC)*2 == len) {
			AL_ARRAY[c].address=PC;
			//DisplayError("%s (%s) found at %08x",AL_ARRAY[c].name,AL_ARRAY[c].libname,AL_ARRAY[c].address);
		}
	}
}

DWORD alGetAddress(const char * searchname) {
	int c;
	for (c=0; c < al_fcn_cnt; c++) {
		if (!AL_ARRAY[c].address) continue;
		if (!strcmp(searchname,AL_ARRAY[c].name)) return AL_ARRAY[c].address;
	}
	return 0;
}

BYTE readbyte(DWORD VAddr) {
	BYTE temp;
	r4300i_LB_VAddr(VAddr,&temp);
	return temp;
}

void dumpseq(DWORD seqaddr) {
	char Directory[255], SaveFile[255];
	OPENFILENAME openfilename;

	if (readbyte(seqaddr) != 'S' || readbyte(seqaddr+1) != '1') {DisplayError("seq header not found at %08x",seqaddr); return;}

	DisplayError(".seq found");
	return;

	memset(&SaveFile, 0, sizeof(SaveFile));
	memset(&openfilename, 0, sizeof(openfilename));

	GetInstantSaveDir( Directory );

	openfilename.lStructSize  = sizeof( openfilename );
	openfilename.hwndOwner    = hMainWindow;
	openfilename.lpstrFilter  = "sequence bank (*.sbk)\0*.seq;";
	openfilename.lpstrFile    = SaveFile;
	openfilename.lpstrInitialDir    = Directory;
	openfilename.nMaxFile     = MAX_PATH;
	openfilename.Flags        = OFN_HIDEREADONLY;

	GetSaveFileName (&openfilename);
}


// ctl size

static unsigned long read32(unsigned long offset)
{
	DWORD d32;
	r4300i_LW_VAddr(offset, &d32);
	return d32;
}

static unsigned int read16(unsigned long offset)
{
	WORD d16;
	r4300i_LH_VAddr(offset, &d16);
	return d16;
}

// needs to be used for percussion as well
static void readinst(unsigned long instoffset, unsigned long ctloffset, unsigned long * maxoffsetp) {
	unsigned long sndarray,c3;
	unsigned int numsnd;
	unsigned long size;
	//printf("\n");

	if (instoffset > *maxoffsetp) *maxoffsetp = instoffset;
	//fseek(infile,offset+ctloffset,SEEK_SET); // seek to instrument

	// skip unneeded data
	//fseek(infile,14,SEEK_CUR); 

	// read number of sounds
	numsnd = read16(instoffset+14);
	//printf("\t\t%d sounds\n",numsnd);

	//copy instrument
	//copydata(ctloffset+offset,offset,16+numsnd*4,infile,outfile);

	// read each sound's data

	sndarray = instoffset+16;

	if (sndarray > *maxoffsetp) *maxoffsetp = sndarray;

	for (c3=0; c3 < numsnd; c3++) {
		unsigned long envelope, keymap, wavetable;
		unsigned long soundoffset;
		soundoffset = ctloffset + read32(sndarray+c3*4);
		if (sndarray+c3*4 > *maxoffsetp) *maxoffsetp = sndarray*c3+4;

		//printf("\t\tsound[%d]=0x%x (0x%x)\n",c3,offset,ctloffset+offset);

		//copy sound
		//copydata(ctloffset+offset,offset,15,infile,outfile);
		if (soundoffset + 15 > *maxoffsetp) *maxoffsetp = soundoffset + 15;

		// read envelope offset
		envelope = read32(soundoffset + 0);

		//copydata(ctloffset+envelope,envelope,14,infile,outfile);
		if (ctloffset+envelope+14 > *maxoffsetp) *maxoffsetp = ctloffset+envelope+14;
		
		// read keymap offset
		//fread(buffer,4,1,infile);
		keymap = read32(soundoffset + 4);

		//copydata(ctloffset+keymap,keymap,6,infile,outfile);
		if (ctloffset+keymap+6 > *maxoffsetp) *maxoffsetp = ctloffset+keymap+6;

		// read wavetable offset (*******)
		wavetable = read32(soundoffset + 8);
		//printf("\t\t\twavetable = 0x%x (0x%x)\n",offset,ctloffset+offset);

		{
			BYTE wavetype;

			// read offset of actual data
			//fread(buffer,4,1,infile);
			//sampledata = read32(buffer);

			// read length of actual data
			//fread(buffer,4,1,infile);
			//samplesize = read32(buffer);

			// read type
			//fread(buffer,1,1,infile);
			r4300i_LB_VAddr(wavetable+8, &wavetype);

			// skip flags+padding
			//fseek(infile,1+2,SEEK_CUR);

			// info
			if (wavetype==0) { // ADPCM
				unsigned long adpcmloop, adpcmbook, size;
				adpcmloop = ctloffset + read32(wavetable+0);
				adpcmbook = ctloffset + read32(wavetable+4);

				// copy wave table
				//copydata(ctloffset+offset,offset,20,infile,outfile);
				if (wavetable + 20 > *maxoffsetp) *maxoffsetp = wavetable + 20;

				// copy loop data
				// (what is the size of ADPCM_STATE?)
				//if (adpcmloop) copydata(ctloffset+adpcmloop,adpcmloop,4*3+0x10,infile,outfile);
				if (adpcmloop!=ctloffset)
				{
					if (adpcmloop+4*3+0x10 > *maxoffsetp) *maxoffsetp = adpcmloop+4*3+0x10;
				}

				size = read32(adpcmbook+4); // read npredictors

				//printf("\t\t\tbook size=%x\n",size);

				// copy book data
				//copydata(ctloffset+adpcmbook,adpcmbook,8+2*size,infile,outfile);
				if (adpcmbook+8+2*size > *maxoffsetp) *maxoffsetp = adpcmbook+8+2*size;
			}
		}

	}
}


unsigned long compute_ctl_size(unsigned long ctloffset)
{
	unsigned long maxoffset = 0, bankarray;
	WORD numbanks, numinst, c;

	// check header
	if (read16(ctloffset) != 0x4231)
	{
		// bad header, don't try to read
		return 8;
	}

	// read number of banks
	numbanks = read16(ctloffset+2);
	//printf("%d banks\n",numbanks);

	// read each bank's data
	bankarray=ctloffset+4;
	maxoffset = bankarray+numbanks*4;
	for (c=0; c < numbanks; c++) {
		unsigned long instarray,c2;
		unsigned long bankoffset, percoffset;

		bankoffset=ctloffset+read32(bankarray+c*4); // read offset of bank structure
		//printf("banks[%d]=0x%x (0x%x)\n",c,offset,ctloffset+offset);

		if (bankoffset+16 > maxoffset) maxoffset = bankoffset+16;

		// read number of instruments
		numinst = read16(bankoffset);
		//printf ("\t%d instruments\n",numinst);

		// read each instrument's data
		//if (offset=read32(buffer)) {printf("percussion yet unimplemented\n"); return 0;}

		instarray = bankoffset+2+6+4;
		percoffset=ctloffset+read32(bankoffset+2+6);

		//printf("\tpercussion=0x%x (0x%x)",offset,ctloffset+offset);
		if (percoffset == ctloffset) ; //printf(" (doesn't exist)\n");
		else readinst(percoffset,ctloffset,&maxoffset);

		for (c2=0; c2 < numinst; c2++) {
			long instoffset;
			instoffset = ctloffset + read32(instarray+c2*4);
			//printf("\tinsts[%d]=0x%x (0x%x)",c2,offset,ctloffset+offset);
			if (instoffset == ctloffset) ; //printf(" (doesn't exist)\n");
			else readinst(instoffset,ctloffset,&maxoffset);
		}

	}

	return maxoffset - ctloffset;
}